import pickle
import os
import numpy as np
import requests

from .aif360 import get_aif360
from .communities import get_communities_and_crimes
from .insurance import get_insurance
from .ihdp import get_ihdp
from .twins import get_twins
from .acs_data import get_acs

dataloaders = {
    'Adult' : lambda : get_aif360('Adult'),
    'Bank' : lambda : get_aif360('Bank'),
    'COMPAS' : lambda : get_aif360('COMPAS'),
    'German' : lambda : get_aif360('German'),
    'MEPS' : lambda : get_aif360('MEPS'),
    'Law School' : lambda : get_aif360('Law School'),
    'Communities' : get_communities_and_crimes,
    'Insurance' : get_insurance,
    'IHDP' : get_ihdp,
    'Twins' : get_twins,
    'ACS': lambda : get_acs('ACSEmployment'),
}

def cache_dataset(name):
    print(f'Caching {name} dataset...')
    url = f'https://raw.githubusercontent.com/rtealwitter/naturalexperiments/main/fairlyuncertain/data/cached/{name}.pkl'
    filename = __file__.replace('__init__.py', f'cached/{name}.pkl')
    r = requests.get(url, timeout=10)
    if r.status_code == 200:
        open(filename, 'wb').write(r.content)
    else:
        output = dataloaders[name]()
        pickle.dump(output, open(filename, 'wb')) 

def read_dataset(name):
    filename = __file__.replace('__init__.py', f'cached/{name}.pkl')
    return pickle.load(open(filename, 'rb'))

def load_instance(name, train_split=.8, max_depth=6, num_boost_round=10, seed=None):
    if seed is not None:
        np.random.seed(seed)
    # Check cached folder exists
    if not os.path.exists(__file__.replace('__init__.py', 'cached')):
        os.makedirs(__file__.replace('__init__.py', 'cached'))
    # Cache dataset if not already cached
    if not os.path.exists(__file__.replace('__init__.py', f'cached/{name}.pkl')):
        cache_dataset(name)    

    real_dataset = read_dataset(name)
    n = real_dataset['n']
    train_indices = np.random.choice(n, int(n*train_split), replace=False)
    test_indices = np.array(list(set(range(n)) - set(train_indices)))
    instance = {
        'name' : name,
        'n': real_dataset['n'],
        'synthetic': False,
        'X': real_dataset['X'],
        'y': real_dataset['y'],
        'group' : real_dataset['group'],        
        'train_indices' : train_indices,
        'group_train' : real_dataset['group'][train_indices],
        'group_test' : real_dataset['group'][test_indices],
        'X_train' : real_dataset['X'][train_indices],
        'X_test' : real_dataset['X'][test_indices],
        'y_train' : real_dataset['y'][train_indices],
        'y_test' : real_dataset['y'][test_indices],
        'max_depth' : max_depth,
        'num_boost_round' : num_boost_round,
    }
    return instance

binary_datasets = ['ACS', 'Adult', 'Bank', 'COMPAS', 'German'] # MEPS requires data agreement
regression_datasets = ['Communities', 'IHDP', 'Insurance', 'Law School', 'Twins']

datasets = binary_datasets + regression_datasets


import pandas as pd

def get_insurance():
    url = "https://raw.githubusercontent.com/stedy/Machine-Learning-with-R-datasets/ff933238d9c30da179b7e3ad4b6ca938d67efc53/insurance.csv"
    data = pd.read_csv(url)
    instance = {}
    # Replace the categorical values with numerical values
    categorical_columns = ['sex', 'region', 'smoker']
    data = pd.get_dummies(data, columns=categorical_columns, drop_first=True)
    # Drop na values
    data = data.dropna()
    instance['data'] = data
    instance['name'] = 'insurance'
    y = data['charges']
    # Normalize the charges
    y = (y - y.min()) / (y.max() - y.min())
    instance['y'] = y.values
    instance['group'] = data['sex_male'].astype(int).values
    instance['X'] = (data.drop(columns=['charges'])).astype(float).values
    instance['n'] = data.shape[0]
    return instance



import pandas as pd

def get_ihdp(nums=[1]):
    # https://github.com/AMLab-Amsterdam/CEVAE/tree/master/datasets/IHDP
    # The IHDP dataset consists of 747 samples with 25 covariates
    # There are 10 IHDP datasets, each with a different treatment effect generated from the same data generating process
    # nums is the number of the 10 IHDP datasets used to build the dataset we use 
    url_prefix = "https://raw.githubusercontent.com/AMLab-Amsterdam/CEVAE/9081f863e24ce21bd34c8d6a41bf0edc7d1b65dd/datasets/IHDP/csv/ihdp_npci_"
    url_suffix = ".csv"
    cols = ['treatment', 'y_factual', 'y_cfactual', 'mu0', 'mu1'] + ['x' + str(i) for i in range(1,26)]

    datasets = []

    for num in nums:
        url = url_prefix + str(num) + url_suffix
        data = pd.read_csv(url, names=cols)
        datasets.append(data)

    data = pd.concat(datasets)
    instance = {}
    instance['data'] = data
    instance['y'] = data['y_factual'].values
    # Normalize the y values
    instance['y'] = (instance['y'] - instance['y'].min()) / (instance['y'].max() - instance['y'].min())
    instance['name'] = 'IHDP'
    instance['group'] = data['x7'].values # This is gender
    instance['X'] = data.drop(['y_factual', 'y_cfactual', 'mu0', 'mu1'], axis=1).values
    instance['n'] = len(instance['y'])

    return instance 


import fairlearn.reductions as red
import xgboost as xgb
import numpy as np
import warnings

moments = {
    'SP' : lambda : red.DemographicParity(),
    'EO' : lambda : red.EqualizedOdds(),
    #'ERP' : lambda : red.ErrorRateParity(),
    'Square' : lambda : red.BoundedGroupLoss(red.SquareLoss(min_val=-np.inf, max_val=np.inf), upper_bound=0.1),
    'Absolute' : lambda : red.BoundedGroupLoss(red.AbsoluteLoss(min_val=-np.inf, max_val=np.inf), upper_bound=0.1),
}

def exponentiated_gradient_algorithm(instance, constraints):
    is_binary = len(np.unique(instance['y_train'])) == 2
    model = xgb.XGBClassifier() if is_binary else xgb.XGBRegressor()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        pred = red.ExponentiatedGradient(
            model, constraints=constraints,
        ).fit(
            instance['X_train'], instance['y_train'], sensitive_features=instance['group_train']
        ).predict(instance['X_test'])
    return {'pred' : pred}

def gridsearch_algorithm(instance, constraints):
    model = red.GridSearch(
        xgb.XGBRegressor(), constraints=constraints,
    )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        model.fit(
            instance['X_train'], instance['y_train'], sensitive_features=instance['group_train']
        )
        pred = model.predict(instance['X_test'])

    return {'pred' : pred}

def exponentiated_gradient(instance, moment_method):
    if moment_method not in moments:
        raise ValueError(f'Invalid moment method. Must be in {moments.keys()}')
    return exponentiated_gradient_algorithm(instance, moments[moment_method]())

def gridsearch(instance, moment_method):
    if moment_method not in moments:
        raise ValueError(f'Invalid moment method. Must be in {moments.keys()}')
    if moment_method in ['SP', 'EO', 'ERP']:
        output = gridsearch_algorithm(instance, moments[moment_method]())
        return {'pred': np.round(output['pred'], 0)}
    return gridsearch_algorithm(instance, moments[moment_method]())



from aif360.datasets import AdultDataset, BankDataset, CompasDataset, GermanDataset, LawSchoolGPADataset, MEPSDataset19

aif360_datasets = {
    'Adult' : {'loader' : AdultDataset, 'protected_attribute_names' : ['sex'], 'privileged_classes' : [['Male']], 'categorical_features' : ['workclass', 'education', 'marital-status', 'occupation', 'relationship', 'native-country', 'race']}, 
    'Bank' : {'loader' : BankDataset, 'protected_attribute_names' : ['age'], 'privileged_classes' : [lambda x : x >= 40], 'categorical_features' : ['job', 'marital', 'education', 'default', 'housing', 'loan', 'contact', 'month', 'day_of_week', 'poutcome']},
    'COMPAS' : {'loader' : CompasDataset, 'protected_attribute_names' : ['sex'], 'privileged_classes' : [['Male']], 'categorical_features' : ['age_cat', 'race', 'c_charge_degree', 'c_charge_desc']},
    'German' : {'loader' : GermanDataset, 'protected_attribute_names' : ['age'], 'privileged_classes' : [lambda x : x >= 25], 'categorical_features': ['status', 'credit_history', 'purpose', 'savings', 'employment', 'other_debtors', 'property', 'installment_plans', 'housing', 'skill_level', 'telephone', 'foreign_worker'], 'features_to_drop':['personal_status', 'sex']},
    'MEPS' : {'loader' : MEPSDataset19, 'protected_attribute_names' : ['RACE'], 'privileged_classes' : [['White']],},
    'Law School' : {'loader' : LawSchoolGPADataset, 'protected_attribute_names' : ['race', 'gender'], 'privileged_classes' : [['white'], ['male']], 'categorical_features' : []},
}

def get_aif360(name):
    if name in ['German', 'MEPS']:
        data, description = aif360_datasets[name]['loader'](
            protected_attribute_names=aif360_datasets[name]['protected_attribute_names'],
            privileged_classes=aif360_datasets[name]['privileged_classes'],
            features_to_drop=aif360_datasets[name].get('features_to_drop', [])
        ).convert_to_dataframe()
    else:
        data, description = aif360_datasets[name]['loader'](
            protected_attribute_names=aif360_datasets[name]['protected_attribute_names'],
            privileged_classes=aif360_datasets[name]['privileged_classes'],
            categorical_features=aif360_datasets[name]['categorical_features'],
        ).convert_to_dataframe()
    # Remove for pickling
    aif360_datasets[name].pop('loader')
    aif360_datasets[name].pop('privileged_classes')
    #aif360_datasets[name]['data'] = data
    aif360_datasets[name]['description'] = description
    label_name = description['label_names'][0]
    protected_attribute_name = description['protected_attribute_names'][0]
    y = data[label_name].values
    X = data.drop(columns=label_name).values
    aif360_datasets[name]['X'] = X
    aif360_datasets[name]['y'] = y
    aif360_datasets[name]['n'] = X.shape[0]
    aif360_datasets[name]['group'] = data[protected_attribute_name].values
    if name in ['German']:
        aif360_datasets[name]['y'] -= 1
    
    return aif360_datasets[name]



from .data import load_instance
from .utils import get_nll
import numpy as np
import os

# # # CALIBRATION # # #

def get_calibration_plot_data(algorithms, datasets):
    results = {}
    for dataset in datasets:
        instance = load_instance(dataset)
        results[dataset] = {'instance' : instance}
        for algorithm_name, algorithm in algorithms.items():
            results[dataset][algorithm_name] = algorithm(instance)
    return results

def load_calibration_results(filename, algorithms, datasets):
    results = {dataset : {algo_name : [] for algo_name in algorithms} for dataset in datasets}
    with open(filename, 'r') as f:
        for line in f:
            saved = eval(line)
            for algo_name in algorithms:
                if algo_name == 'dataset': continue
                results[saved['dataset']][algo_name].append(saved[algo_name])
    return results

def get_calibration_table_data(is_binary, algorithms, datasets, num_runs=10, folder=None):
    if folder is not None:
        type = 'binary' if is_binary else 'regression'
        filename = folder + f'/calibration_{type}.csv'
        if os.path.exists(filename):
            return load_calibration_results(filename, algorithms, datasets) 
    results = {}

    for dataset in datasets:
        results[dataset] = {algo_name : [] for algo_name in algorithms}
        for _ in range(num_runs):
            instance = load_instance(dataset)
            for algo_name, algorithm in algorithms.items():
                output = algorithm(instance)
                nll = get_nll(output['pred'], output['std'], instance['y_test'], is_binary).mean()
                results[dataset][algo_name].append(nll)
            if folder is not None:
                saved = {algo_name : results[dataset][algo_name][-1] for algo_name in algorithms}
                saved['dataset'] = dataset
                with open(filename, 'a') as f:
                    if 'nan' not in str(saved):
                        f.write(str(saved) + '\n')

    if folder is not None:
        results = load_calibration_results(filename, algorithms, datasets)

    return results

# # # CONSISTENCY # # #

def load_consistency_results(filename, algorithms):
    results = {name: {} for name in algorithms}
    if not os.path.exists(filename):
        return results    
    with open(filename, 'r') as f:
        for line in f:
            saved = eval(line)
            max_depth = saved['max_depth']
            for name in saved:
                if name == 'max_depth': continue
                if max_depth not in results[name]:
                    results[name][max_depth] = []
                results[name][max_depth] += [saved[name]]
    return results

def get_consistency_data(dataset, algorithms, max_depths):
    filename = f'cached/consistency_{dataset}.csv'
    instance = load_instance(dataset)
    for max_depth in max_depths:
        
        instance['max_depth'] = max_depth
        saved = {'max_depth': max_depth}
        for algo_name, algorithm in algorithms.items():
            output = algorithm(instance)
            saved[algo_name] = list(output['std'])
        if len(saved) > 1:
            with open(filename, 'a') as f:
                f.write(str(saved) + '\n')
    
    return load_consistency_results(filename, algorithms)

# # # BINARY FAIRNESS # # #

def compute_binary_fairness(results, algorithms, metrics): 
    metric_values = {metric : {algo_name : [] for algo_name in algorithms} for metric in metrics}
    metric_values['Included \%'] = {algo_name : [] for algo_name in algorithms}

    for num_run in results:
        instance = results[num_run]['instance']
        y = instance['y_test']
        group = instance['group_test'] == 1
        for algo_name in algorithms:
            pred = results[num_run][algo_name]['pred']
            # Gridsearch returns floats
            pred = pred.clip(0, 1)
            pred = np.round(pred)
            assert np.all(np.unique(pred)==np.array([0,1])), 'Predictions are not binary'
            # No abstaining
            if 'std' not in results[num_run][algo_name]:
                for metric in metrics:
                    val = metrics[metric](pred, y, group, run_checks=False)
                    metric_values[metric][algo_name].append(val)
                metric_values['Included \%'][algo_name].append(100)
                continue
            # Abstaining at best percentile for SP
            percentiles = np.arange(75, 101, 1)
            std = results[num_run][algo_name]['std'] 
            best_val = np.inf
            for percentile in percentiles:
                include = std <= np.percentile(std, percentile)
                val = 0
                for metric in metrics:
                    normalization = metrics[metric](results[num_run]['Baseline']['pred'], y, group, run_checks=False)
                    metric_val = metrics[metric](pred[include], y[include], group[include], run_checks=False)
                    val += metric_val / normalization
                if val < best_val:
                    best_val = val
                    best_percentile = percentile
            include = std <= np.percentile(std, best_percentile) 
            for metric in metrics:
                metric_val = metrics[metric](pred[include], y[include], group[include], run_checks=False)
                metric_values[metric][algo_name].append(metric_val)
            metric_values['Included \%'][algo_name].append(int(100 * include.mean()))

    return metric_values


import numpy as np

def binomial_nll(instance, Model):
    model = Model('classifier', instance)
    model.fit(instance['X_train'], instance['y_train'])
    p = model.predict_proba(instance['X_test'])[:,1]
    pred = model.predict(instance['X_test'])
    std = np.sqrt(p * (1-p))
    return {'pred': pred, 'std': std, 'p': p}


import scipy
import numpy as np
import warnings

# # # FORMATTING # # #

def fancy_round(x, precision=3):
    return float(np.format_float_positional(x, precision=precision, unique=False, fractional=False, trim='k'))

def print_table(results_all, filename, include_var=True, lower_is_better=True):
    with open(filename, 'w') as f:
        col_types = 'l' + 'c' * len(results_all.keys())

        f.write('\\begin{tabular} {' + col_types + '}\n')
        f.write('\\toprule\n')

        col_names = ['Approach'] + list(results_all.keys())
        col_names_bold = ['\\textbf{' + col_name + '}' for col_name in col_names]

        f.write(' & '.join(col_names_bold) + ' \\\\ \\midrule\n')

        current_round = lambda x: round(fancy_round(x, 3), 3)
        
        mean_cols = {}
        for dataset in results_all:
            mean_col = []
            for algo_name in results_all[dataset]:
                mean_col.append(current_round(np.mean(results_all[dataset][algo_name])))
            if 'Included' in dataset:
                lower_is_better = not lower_is_better
            mean_col_sorted = sorted(mean_col, reverse=not lower_is_better)
            mean_cols[dataset] = mean_col_sorted

        first_key = list(results_all.keys())[0]

        for algo_name in results_all[first_key]:
            to_print = [algo_name]
            for dataset in results_all:
                mean_val = current_round(np.mean(results_all[dataset][algo_name]))
                std_val = current_round(np.std(results_all[dataset][algo_name]))
                color = ''
                if mean_val == mean_cols[dataset][0]:
                    color = '\\cellcolor{gold!30}'
                elif mean_val == mean_cols[dataset][1]:
                    color = '\\cellcolor{silver!30}'
                elif mean_val == mean_cols[dataset][2]:
                    color = '\\cellcolor{bronze!30}'              
                mean_val_str = f'{color}{mean_val}'
                
                if include_var:
                    to_print.append(f'{mean_val_str} $\\pm$ {std_val}')
                else:
                    to_print.append(mean_val_str)

            f.write(' & '.join(to_print) + ' \\\\ \n')

        f.write('\\bottomrule\n')
        f.write('\\end{tabular}\n')

def get_consistency_table(results, datasets, algorithms):
    table = {}
    for dataset in datasets:
        table[dataset] = {algo_name : [] for algo_name in algorithms}
        for algo_name in algorithms.keys():            
            max_depths = sorted(results[dataset][algo_name].keys())
            for i in range(len(results[dataset][algo_name][max_depths[0]])):
                std = np.array([results[dataset][algo_name][max_depth][i] for max_depth in max_depths])
                max_std = (np.std(std, axis=0)).max()
                table[dataset][algo_name].append(max_std)
    
    return table

def bold_maxes(rows):
    names = rows[:,0]
    X = np.array(rows[:,1:], dtype=float)
    col_sorted = []
    for col in X.T:
        col_sorted.append(sorted(col))

    for j, row in enumerate(X):
        to_print = [names[j]]
        for i, value in enumerate(row):
            rounded = round(value, 3)
            val = str(rounded)
            if value == col_sorted[i][0]:
                val = f'\\textbf{{{rounded}}}'
            elif value == col_sorted[i][1]:
                val = '\\textbf{\\textit{' + str(rounded) + '}}'
            elif value == col_sorted[i][2]:
                val = '\\textbf{\\underline{' + str(rounded) + '}}'
            to_print.append(val)
        print(' & '.join(to_print) + ' \\\\ \\hline')

# # # CALIBRATION # # # 

def get_binary_nll(pred, std, y):
    std = np.clip(std, 0, .5)
    ptilde_a = (1 + np.sqrt(1 - 4 * std**2))/2
    ptilde_b = (1 - np.sqrt(1 - 4 * std**2))/2
    ptilde = (pred > .5) * ptilde_a + (pred <= .5) * ptilde_b
    ptilde = np.clip(ptilde, 0.001, .999)
    return -1 * (np.log(ptilde) * y + np.log(1-ptilde) * (1-y))

def get_regression_nll(pred, std, y):
    return np.log(std * np.sqrt(2 * np.pi)) + (y - pred)**2 / (2 * std**2)

def get_nll(pred, std, y, is_binary):
    if is_binary:
        return get_binary_nll(pred, std, y)
    else:
        return get_regression_nll(pred, std, y)
    
# # # BINARY FAIRNESS # # #
from fairlearn.metrics import demographic_parity_difference, equalized_odds_difference

def get_diffs(lst):
    max_diff = 0
    for i in range(len(lst)):
        for j in range(len(lst)):
            diff = np.abs(lst[i] - lst[j])
            max_diff = max(max_diff, diff)
    return max_diff

def binary_statistical_parity(pred, y, group, run_checks=True):
    mean_preds = [] 
    for g in np.unique(group):
        include = group == g
        mean_pred = pred[include].mean()
        mean_preds.append(mean_pred)
    val1 = get_diffs(mean_preds)

    if not run_checks:
        return val1
    print('running checks')
    
    val2 = demographic_parity_difference(y, pred, sensitive_features=group)

    error_msg = 'Error in SP...' + str(val1) + '!=' + str(val2)
    assert val1 == val2, error_msg

    return val1


def binary_equalized_odds(pred, y, group, run_checks=True):
    # True positive rate difference
    # False positive rate difference
    # Return the maximum of the two
    tprs, fprs = [], []
    for g in np.unique(group):
        include = group == g
        tpr = np.mean(pred[include][y[include] == 1])
        fpr = np.mean(pred[include][y[include] == 0])
        if np.isnan(tpr): tpr = 0
        if np.isnan(fpr): fpr = 0
        tprs += [tpr]
        fprs += [fpr]
    tpr_diff = get_diffs(tprs)
    fpr_diff = get_diffs(fprs)

    val1 = max(tpr_diff, fpr_diff)

    if not run_checks:
        return val1
    
    print('running checks')

    val2 = equalized_odds_difference(y, pred, sensitive_features=group)

    error_msg = 'Error in EO...' + str(val1) + '!=' + str(val2)

    assert val1 == val2, error_msg

    return val1

def binary_error_rate(pred, y, group, run_checks=True):
    return np.mean(pred != y)
    
binary_metrics = {
    'Error Rate': binary_error_rate,
    'Statistical Parity': binary_statistical_parity, 
    'Equalized Odds': binary_equalized_odds,
    #'Equal Opportunity': binary_equal_opportunity
}

# # # REGRESSION FAIRNESS # # #

def compute_cdf(output, z, group_indicator=None):
    z = z.reshape(-1, 1)
    if group_indicator is None:
        group_indicator = np.ones(len(output['pred']), dtype=bool)

    mus = output['pred'][group_indicator].squeeze()
    if 'std' in output:
        sigma = output['std'][group_indicator]
        cdf = scipy.stats.norm.cdf(z, mus, sigma).mean(axis=1)
    else:
        cdf = (mus < z).mean(axis=1)
    return cdf

def general_sp(cdfs):
    max_diff = 0
    for i in range(cdfs.shape[0]):
        for j in range(cdfs.shape[0]):
            if i == j:
                continue
            diff = np.abs(cdfs[i] - cdfs[j]).max()
            max_diff = max(max_diff, diff)
    return max_diff

def get_regression_statistical_parity(y, group, output):
    z = np.linspace(min(y), max(y), 1000)
    cdfs = []
    for g in np.unique(group):
        cdf = compute_cdf(output, z, group == g)
        cdfs += [cdf]
    return general_sp(np.array(cdfs))



import pandas as pd

def get_twins():
    # https://github.com/AMLab-Amsterdam/CEVAE/tree/master/datasets/TWINS
    # Twins dataset
    url_weight = "https://raw.githubusercontent.com/AMLab-Amsterdam/CEVAE/9081f863e24ce21bd34c8d6a41bf0edc7d1b65dd/datasets/TWINS/twin_pairs_T_3years_samesex.csv"

    url_X = "https://raw.githubusercontent.com/AMLab-Amsterdam/CEVAE/9081f863e24ce21bd34c8d6a41bf0edc7d1b65dd/datasets/TWINS/twin_pairs_X_3years_samesex.csv"

    url_y = "https://raw.githubusercontent.com/AMLab-Amsterdam/CEVAE/9081f863e24ce21bd34c8d6a41bf0edc7d1b65dd/datasets/TWINS/twin_pairs_Y_3years_samesex.csv"

    weight = pd.read_csv(url_weight) # birth weight
    X = pd.read_csv(url_X) # covariates
    y = pd.read_csv(url_y) # mortality (binary)

    data = pd.concat([weight, X, y], axis=1).drop(columns=['Unnamed: 0', 'Unnamed: 0.1'])
    # Drop data where nprevisitq is nan
    data = data.dropna(subset=['nprevistq'])
    # Drop columns with missing values
    data = data.dropna(axis=1)
    data['sensitive'] = (data['crace'] == 1).astype(int)

    instance = {}

    instance['data'] = data
    instance['group'] = data['sensitive'].values
    instance['y'] = data['nprevistq'].values
    # Normalize the y values
    instance['y'] = (instance['y'] - instance['y'].min()) / (instance['y'].max() - instance['y'].min())
    instance['X'] = data.drop(columns=['nprevistq']).values.astype(float)
    instance['n'] = instance['X'].shape[0]

    return instance



def normal_nll_wrap(instance, loss_name, Model):
    model = Model(loss_name, instance)
    model.fit(instance['X_train'], instance['y_train'])
    preds = model.predict(instance['X_test'])
    mu, sigma = preds[:,0], preds[:,1]
    return {'pred' : mu, 'std' : sigma}

def normal_nll(instance, Model):
    return normal_nll_wrap(instance, 'Normal NLL', Model)

def beta_nll(instance, Model):
    return normal_nll_wrap(instance, r'$\beta$-NLL', Model)

def faithful_nll(instance, Model):
    return normal_nll_wrap(instance, 'Faithful NLL', Model)


from folktables import ACSDataSource, ACSEmployment, ACSIncome, ACSPublicCoverage

import pandas as pd
import numpy as np

data_types = {
    "AGEP": {
        "type":"ordinal"
    },
    "ANC": {
        "type":"categorical"
    },
    "CIT": {
        "type":"categorical"
    },
    "COW": {
        "type":"categorical"
    },
    "DEAR": {
        "type":"categorical"
    },
    "DEYE": {
        "type":"categorical"
    },
    "DIS": {
        "type":"categorical"
    },
    "DREM": {
        "type":"categorical"
    },
    "ESP": {
        "type":"categorical"
    },
    "ESR": {
        "type":"categorical"
    },
    "FER": {
        "type":"categorical"
    },
    "JWTR": {
        "type":"categorical"
    },
    "MAR": {
        "type":"categorical"
    },
    "MIG": {
        "type":"categorical"
    },
    "MIL": {
        "type":"categorical"
    },
    "NATIVITY": {
        "type":"categorical"
    },
    "OCCP": {
        "type":"categorical"
    },
    "PINCP": {
        "type":"categorical"
    },
    "POBP": {
        "type":"categorical"
    },
    "POVPIP": {
        "type":"numerical"
    },
    "POWPUMA": {
        "type":"categorical"
    },
    "PUMA": {
        "type":"categorical"
    },
    "RAC1P": {
        "type":"categorical"
    },
    "GCL": {
        "type":"categorical"
    },
    "RELP": {
        "type":"categorical"
    },
    "SCHL": {
        "type":"categorical"
    },
    "SEX": {
        "type":"categorical"
    },
    "ST": {
        "type":"categorical"
    },
    "WKHP": {
        "type":"ordinal"
    },
    "PUBCOV": {
        "type":"categorical"
    },
    "JWMNP": {
        "type":"categorical"
    },
}

def get_categorical_features(scenario, data_types):
    categorical_features = []
    for feature in scenario._features:
        if data_types[feature]['type'] == 'categorical':
            categorical_features.append(feature)
    return categorical_features

ACS_Scenarios = {
    "ACSEmployment": {'loader': ACSEmployment, 
                      'categorical_features' : get_categorical_features(ACSEmployment, data_types),
                      'sensitive_features': [1, 2]
                      },
    "ACSIncome": {'loader': ACSIncome, 
                      'categorical_features' : get_categorical_features(ACSIncome, data_types),
                      'sensitive_features': [1, 2]
                 },
    "ACSPublicCoverage": {'loader': ACSPublicCoverage, 
                          'categorical_features' : get_categorical_features(ACSPublicCoverage, data_types),
                          'sensitive_features': [1, 2]
                         }
}

def return_acs_data_scenario(acs_data, scenario="ACSEmployment", verbose=False):
    scenario = ACS_Scenarios[scenario]['loader']
    features, label, group = scenario.df_to_numpy(acs_data)
    if verbose:
        print(features, label, group)

    np_all_data = np.c_[features,label]

    df = pd.DataFrame(np_all_data, columns = scenario._features + [scenario._target])
    features = pd.DataFrame(features, columns = scenario._features)
    target = pd.DataFrame(label, columns = [scenario._target])
    group = pd.DataFrame(group, columns = [scenario._group])

    return df, features, target, group

def get_acs(name, states=['NY']):
    data_source = ACSDataSource(survey_year='2018', horizon='1-Year', survey='person')
    acs_data = data_source.get_data(states=states, download=True)
    _, features, target, group = return_acs_data_scenario(acs_data, scenario=name)

    # label_name = target.columns
    protected_attribute_name = group.columns
    y = target
    y = y.astype(int)
    X = features

    criteria = group[protected_attribute_name[0]].isin(ACS_Scenarios[name]['sensitive_features'])
    group_filtered = group[criteria]
    group_filtered = group_filtered - 1
    X = X[criteria]
    y = y[criteria]
    
    indices_subsample = np.random.choice(X.shape[0], int(X.shape[0]*0.1), replace=False)
    X = X.iloc[indices_subsample]
    y = y.iloc[indices_subsample]
    group_filtered = group_filtered.iloc[indices_subsample]

    ACS_Scenarios[name].pop('loader')
    ACS_Scenarios[name]['X'] = X.values
    ACS_Scenarios[name]['y'] = y.values.ravel()
    ACS_Scenarios[name]['n'] = X.shape[0]
    ACS_Scenarios[name]['group'] = group_filtered.values.ravel()
    ACS_Scenarios[name]['protected_attribute_names'] = protected_attribute_name
    
    return ACS_Scenarios[name]

    


  


import numpy as np

rng = np.random.default_rng()

def ensemble(instance, Model, num_ensemble=10):
    models = []
    # Check if instance['y_train'] is binary
    is_binary = len(np.unique(instance['y_train'])) == 2
    n = len(instance['X_train'])
    for i in range(num_ensemble):
        indices = rng.choice(n, n, replace=True)
        X = instance['X_train'][indices]
        y = instance['y_train'][indices]
        model = Model('classifier' if is_binary else 'regression', instance)
        model.fit(X, y)
        models.append(model)
    
    model_predict = lambda model, X : model.predict(X) if not is_binary else model.predict_proba(X)[:,1]
    
    def predict(X):
        preds = np.array([model_predict(model, X) for model in models]).T
        return np.mean(preds, axis=1), np.std(preds, axis=1)
    
    mu, sigma = predict(instance['X_test'])
    return {'pred': mu, 'std': sigma}


import torch
import torch.nn.functional as F
from torch.utils.data import DataLoader, TensorDataset
from torch.cuda.amp import autocast, GradScaler

import pandas as pd
import numpy as np
from tqdm import tqdm

from sklearn.model_selection import RandomizedSearchCV, train_test_split
from sklearn.base import BaseEstimator
from sklearn.metrics import accuracy_score

from tab_transformer_pytorch import FTTransformer
import os

class FTTransformerModel(BaseEstimator):
    def __init__(self,
                 categories,
                 num_continuous,
                 dim = 32,                           # dimension, paper set at 32
                 dim_out = 2,                        # binary prediction, but could be anything
                 depth = 6,                          # depth, paper recommended 6
                 heads = 8,                          # heads, paper recommends 8
                 attn_dropout = 0.1,                 # post-attention dropout
                 ff_dropout = 0.1,                   # feed forward dropout
                 batch_size=1,
                 num_epochs=50,
                 lr=3e-4,
                 load_best_model_when_trained=True,
                 verbose=False,
                 loss_fn=None,
                 squeeze_output=False,
                 model_save_path='model_cache/best_model.pth'):  # Add model_save_path parameter
        super(FTTransformerModel, self).__init__()

        self.verbose = verbose
        self.dim = dim
        self.dim_out = dim_out
        self.depth = depth
        self.heads = heads
        self.attn_dropout = attn_dropout
        self.ff_dropout = ff_dropout

        self.batch_size = batch_size
        self.num_epochs = num_epochs
        self.lr = lr
        self.load_best_model_when_trained = load_best_model_when_trained
        self.best_model_dict = None
        self.loss_fn = loss_fn
        self.squeeze_output = squeeze_output
        self.model_save_path = model_save_path  # Initialize model save path

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.device = device

        self.model = self.build_model(
            categories,
            num_continuous,
            dim,
            dim_out,
            depth,
            heads,
            attn_dropout,
            ff_dropout,
        ).to(self.device)

    def build_model(self, categories, num_continuous, dim, dim_out, depth, heads, attn_dropout, ff_dropout):
        model = FTTransformer(
            categories = categories,
            num_continuous = num_continuous,
            dim = dim,
            dim_out = dim_out,
            depth = depth,
            heads = heads,
            attn_dropout = attn_dropout,
            ff_dropout = ff_dropout
        )

        return model

    def fit(self,
            X_cat_train,
            X_cont_train,
            y_train,
            patience=2,
            accumulation_steps=1):
        
        torch.autocast(device_type='cuda', dtype=torch.float16)

        X_cat_train, X_cont_train, y_train = X_cat_train.to(self.device), X_cont_train.to(self.device), y_train.to(self.device)
        
        X_cat_train, X_cat_val, X_cont_train, X_cont_val, y_train, y_val = train_test_split(X_cat_train,
                                                                                            X_cont_train,
                                                                                            y_train,
                                                                                            test_size=0.1,
                                                                                            random_state=42)
        
        train_dataset = TensorDataset(X_cat_train, X_cont_train, y_train)
        train_loader = DataLoader(train_dataset, batch_size=self.batch_size, shuffle=True)
        
        val_dataset = TensorDataset(X_cat_val, X_cont_val, y_val)
        val_loader = DataLoader(val_dataset, batch_size=self.batch_size, shuffle=False)

        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        scaler = GradScaler()  # Initialize the gradient scaler for mixed precision

        best_accuracy = 0.0
        best_loss = np.inf
        epochs_without_improvement = 0

        for epoch in range(self.num_epochs):
            self.model.train()
            if self.verbose:
                batch_progress = tqdm(train_loader, desc=f'Epoch {epoch+1}/{self.num_epochs}', unit='batch')
            else:
                batch_progress = train_loader

            optimizer.zero_grad()
            for i, (X_batch_cat, X_batch_cont, y_batch) in enumerate(batch_progress):
                X_batch_cat, X_batch_cont, y_batch = X_batch_cat.to(self.device), X_batch_cont.to(self.device), y_batch.to(self.device)

                with autocast(): 
                    output = self.model(X_batch_cat, X_batch_cont)
                    y_batch = y_batch.type(torch.LongTensor).to(self.device)
                    if self.squeeze_output:
                        output = output.squeeze()
                    loss = self.loss_fn(output, y_batch.float())

                scaler.scale(loss).backward() 

                if (i + 1) % accumulation_steps == 0:
                    scaler.step(optimizer) 
                    scaler.update() 
                    optimizer.zero_grad()

                if self.verbose:
                    batch_progress.set_description(f'Epoch {epoch+1}/{self.num_epochs} Loss: {loss.item():.4f}')
                    batch_progress.refresh()

            if (i + 1) % accumulation_steps != 0:
                scaler.step(optimizer)
                scaler.update()
                optimizer.zero_grad()

            self.model.eval()
            val_loss = 0.0
            val_correct = 0
            val_total = 0
            with torch.no_grad():
                for X_batch_cat, X_batch_cont, y_batch in val_loader:
                    X_batch_cat, X_batch_cont, y_batch = X_batch_cat.to(self.device), X_batch_cont.to(self.device), y_batch.to(self.device)
                    with autocast():
                        val_output = self.model(X_batch_cat, X_batch_cont)
                        if self.squeeze_output:
                            val_output = val_output.squeeze()
                        y_batch = y_batch.type(torch.LongTensor).to(self.device)
                        batch_val_loss = self.loss_fn(val_output, y_batch.float()).item()
                        val_loss += batch_val_loss * y_batch.size(0)
                        if self.squeeze_output:
                            predict_proba = val_output
                        else:
                            predict_proba = val_output[:, 0]
                        val_preds_binary = (predict_proba > 0.5).cpu().numpy().astype(int)
                        val_correct += (val_preds_binary == y_batch.cpu().numpy()).sum()
                        val_total += y_batch.size(0)

            val_loss /= val_total
            val_accuracy = val_correct / val_total

            if val_loss < best_loss:
                best_loss = val_loss
                best_accuracy = val_accuracy
                torch.save(self.model.state_dict(), self.model_save_path)
                if self.verbose:
                    print(f"Val loss - new best: {best_loss}")
                    print(f"Val accuracy - new best: {best_accuracy}")
                epochs_without_improvement = 0
            else:
                epochs_without_improvement += 1

            if self.verbose:
                print(f"Epoch {epoch+1}, Validation Loss: {val_loss}, Epochs without improvement: {epochs_without_improvement}/{patience}")

            if epochs_without_improvement >= patience:
                print(f"Stopping early at epoch {epoch+1}. No improvement in validation loss for {patience} consecutive epochs.")
                print(f"Epoch {epoch+1}, Validation Loss: {val_loss}, Epochs without improvement: {epochs_without_improvement}/{patience}")
                print(f"Val accuracy - new best: {best_accuracy}")
                break

            with torch.no_grad():
                torch.cuda.empty_cache()

        if self.load_best_model_when_trained and os.path.exists(self.model_save_path):
            self.model.load_state_dict(torch.load(self.model_save_path))

    def load_best_model(self):
        """Loads the best model parameters."""
        if os.path.exists(self.model_save_path):
            self.model.load_state_dict(torch.load(self.model_save_path))
        else:
            print("No best model saved. Please run the training first.")

    def predict(self, X_cat_test, X_cont_test, binary=False, threshold=0.5):
        with torch.no_grad():
            X_cat_test, X_cont_test = X_cat_test.to(self.device), X_cont_test.to(self.device)
            predictions = self.model(X_cat_test, X_cont_test)
            if self.squeeze_output:
                predictions = predictions.squeeze()
            proba = predictions.softmax(dim=-1)
            if binary:
                predictions = (proba[:, 0] > threshold).cpu().numpy().astype(int)
            else:
                predictions = predictions.cpu().numpy()
            return predictions
        
    def predict_proba(self, X_cat_test, X_cont_test):
        with torch.no_grad():
            X_cat_test, X_cont_test = X_cat_test.to(self.device), X_cont_test.to(self.device)
            predictions = self.model(X_cat_test, X_cont_test)
            if self.squeeze_output:
                predictions = predictions.squeeze()
            proba = predictions.softmax(dim=-1)
            return proba.cpu().numpy()



import torch
import numpy as np
import scipy
from torch.utils.data import DataLoader, Dataset

seed = 0
rng = np.random.default_rng(seed=seed)

# Neural Networks
class SelectiveSquare(torch.nn.Module): 
    def __init__(self): 
        super().__init__()
        return 

    def forward(self,x): 
        out = torch.zeros_like(x)
        out[:,0] = x[:,0]
        # Square the second variable (std is non-negative)
        out[:,1] = x[:,1]**2
        return out

class SeparateGradient(torch.nn.Module):
    def __init__(self, intermediate_size=100, num_layers=3, activation = torch.nn.ReLU):
        super().__init__()
        mean_layers = []
        std_layers = []
        for _ in range(num_layers):
            mean_layers.append(torch.nn.Linear(intermediate_size, intermediate_size))
            mean_layers.append(activation())
            std_layers.append(torch.nn.Linear(intermediate_size, intermediate_size))
            std_layers.append(activation())
        mean_layers.append(torch.nn.Linear(intermediate_size, 1))
        std_layers.append(torch.nn.Linear(intermediate_size, 1))
        self.mean_layers = torch.nn.Sequential(*mean_layers)
        self.std_layers = torch.nn.Sequential(*std_layers)
        return

    def forward(self,x):
        mean = self.mean_layers(x).squeeze()
        std = self.std_layers(x.detach()).squeeze()
        out = torch.zeros((x.shape[0], 2))
        out[:,0] = mean
        out[:,1] = std **2
        return out
    
def build_separate_model(input_size, num_layers, activation = torch.nn.ReLU, dropout=0, intermediate_size=100):
    layers = []
    layers.append(torch.nn.Linear(input_size, intermediate_size))
    layers.append(activation())
    for _ in range(num_layers//2):
        layers.append(torch.nn.Dropout(dropout))
        layers.append(torch.nn.Linear(intermediate_size, intermediate_size))
        layers.append(activation())
    layers.append(SeparateGradient(intermediate_size, num_layers//2, activation))
    return torch.nn.Sequential(*layers)

def build_model(input_size, output_size, num_layers, activation = torch.nn.ReLU, dropout=0, intermediate_size=100):
    layers = []
    layers.append(torch.nn.Linear(input_size, intermediate_size))
    layers.append(activation())
    for _ in range(num_layers):
        layers.append(torch.nn.Dropout(dropout))
        layers.append(torch.nn.Linear(intermediate_size, intermediate_size))
        layers.append(activation())
    layers.append(torch.nn.Linear(intermediate_size, output_size))
    if output_size == 2:
        layers.append(SelectiveSquare())
    else:
        # Logits
        layers.append(torch.nn.Sigmoid())
    model = torch.nn.Sequential(*layers)
    model.train()
    return model

class CustomDataset(Dataset):
    def __init__(self, X, y):
        self.X = torch.tensor(X, dtype=torch.float32)
        self.y = torch.tensor(y, dtype=torch.float32)

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        return self.X[idx], self.y[idx]

def train(X, y, model, loss_fn, num_epochs, num_batches, learning_rate):
    n = X.shape[0]
    optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)
    X_tensor = torch.tensor(X, dtype=torch.float32)
    y_tensor = torch.tensor(y, dtype=torch.float32)
    # Train the model
    batch_size = n // num_batches
    for t in range(num_epochs):
        # Shuffle the indices
        indices = rng.choice(n, n, replace=False)
        for batch in range(0, n, batch_size):
            # Compute the batch indices
            batch_indices = indices[batch:batch+batch_size]
            X_batch = X_tensor[batch_indices]
            y_batch = y_tensor[batch_indices]
            # Forward pass: compute predicted y
            pred = model(X_batch)
            loss = loss_fn(pred.squeeze(), y_batch)
            # Backward pass: zero, backpropagate and step
            model.zero_grad()
            loss.backward()
            optimizer.step()

    easy_input_model = lambda x: model(torch.tensor(x, dtype=torch.float32)).detach().numpy()
    return easy_input_model

def normal_nll(pred, y_batch):
    mean_pred, std_pred = pred[:,0], pred[:,1]
    std_pred += 1e-6 # Avoid division by zero

    # Compute loss
    loss= (mean_pred - y_batch)**2 / (2*std_pred**2) + torch.log((std_pred)**2)

    return loss.mean()

def beta_nll(pred, y_batch):
    beta = .5
    mean_pred, std_pred = pred[:,0], pred[:,1]
    weighting = (std_pred**(2*beta)).detach()


    # Compute loss
    loss= weighting*((mean_pred - y_batch)**2 / (2*std_pred) + torch.log((std_pred)**2))

    return loss.mean()

def faithful_nll(pred, y_batch):
    mean_pred, std_pred = pred[:,0], pred[:,1]

    std_pred += 1e-6 # Avoid division by zero

    loss_mean = (mean_pred - y_batch)**2
    
    loss_std = (mean_pred.detach() - y_batch)**2 / (2*std_pred**2) + torch.log((std_pred)**2)

    # Compute loss
    loss = loss_mean + loss_std

    return loss.mean()

def binomial(pred, y_batch):
    return torch.nn.functional.binary_cross_entropy_with_logits(pred, y_batch)

def squared_error(pred, y_batch):
    return torch.nn.functional.mse_loss(pred, y_batch)

def nn_train(loss_name, nn_instance):
    if 'num_batches' not in nn_instance:
        nn_instance['num_batches'] = 5
    input_size = nn_instance['X_train'].shape[1]
    dropout = 0 if 'Dropout' not in loss_name else 0.2
    output_size = 1 if 'BCE' in loss_name else 2

    if 'Faithful' in loss_name:
        model = build_separate_model(input_size, nn_instance['num_layers'], dropout=dropout)
    else:
        model = build_model(input_size, output_size, nn_instance['num_layers'], dropout=dropout)

    model = train(nn_instance['X_train'], nn_instance['y_train'], model, nn_losses[loss_name], nn_instance['num_epochs'], nn_instance['num_batches'])

    return model

losses = {
    'Normal NLL' : normal_nll,
    'Faithful NLL' : faithful_nll,
    r'$\beta$-NLL' : beta_nll,
    'Logistic' : binomial,
    'Squared Error' : squared_error
}

class Model:
    def __init__(self, loss_name=None, instance=None):
        if instance is None: instance = {}

        # Model parameters
        depth = 6 if 'depth' not in instance else instance['depth']
        dropout = 0 if 'dropout' not in instance else instance['dropout']
        input_size = instance['X_train'].shape[1]
        output_size = 1 if loss_name in ['Logistic', 'Squared Error'] else 2

        if 'Faithful' in loss_name:
            self.model = build_separate_model(input_size, depth, dropout=dropout)
        else:
            self.model = build_model(input_size, output_size, depth, dropout=dropout)
        
        # Training parameters
        self.epochs = 100 if 'epochs' not in instance else instance['epochs']
        self.learning_rate = 0.01 if 'learning_rate' not in instance else instance['learning_rate']
        self.num_batches = 5 if 'num_batches' not in instance else instance['num_batches']
        self.loss = losses[loss_name]
        self.loss_name = loss_name
    
    def fit(self, X, y):
        self.model = train(X, y, self.model, self.loss, self.epochs, self.num_batches, self.learning_rate)
            
    def predict(self, X):
        if self.loss_name == 'Logistic':
            return self.model(X).round()
        return self.model(X)
 
    def predict_proba(self, X):
        return self.model(X)



import matplotlib.pyplot as plt
import numpy as np
from .utils import compute_cdf

plt.rcParams.update({'font.size': 16})

linestyles = ['-', '--', '-.', ':', (0, (3, 5, 1, 5)), (1, (3, 5, 1, 5)), (2, (3, 5, 1, 5))]


# # # CALIBRATION # # #

def plot_calibration(results, is_binary, algorithms, datasets, folder='figures'):
    fig, axs = plt.subplots(1, len(datasets), figsize=(20, 3))

    markers = ['o', 's', 'v', 'D', 'P', 'X', 'H']

    for i, dataset in enumerate(datasets):
        instance = results[dataset]['instance']
        y = instance['y_test']
        for j, algo_name in enumerate(algorithms):
            std = results[dataset][algo_name]['std']
            if not is_binary:
                mean = results[dataset][algo_name]['pred']
            sigma_bins = np.linspace(np.percentile(std, 0), np.percentile(std, 100), 10)
            pred_std, empirical_std = [], []
            for k in range(len(sigma_bins)-1):
                criteria = (std >= sigma_bins[k]) & (std < sigma_bins[k+1])
                if np.sum(criteria) < 20: continue
                pred_std.append(np.mean(std[criteria]))
                if not is_binary:
                    y_centered = y[criteria] - mean[criteria]
                    empirical_std.append(np.std(y_centered))
                else:
                    empirical_std.append(np.std(y[criteria]))
            axs[i].plot(pred_std, empirical_std, marker=markers[j], label=algo_name)
            
        # get xlim and ylim
        min_val = min(axs[i].get_xlim()[0], axs[i].get_ylim()[0])
        max_val = max(axs[i].get_xlim()[1], axs[i].get_ylim()[1])
        axs[i].plot([min_val, max_val], [min_val, max_val], alpha=.5, linestyle='--', color='black')
        axs[i].set_title(dataset)


        axs[i].set_xlabel('Predicted')
        if i == 0: axs[i].set_ylabel('Empirical')
        # Decrease size of tick font
        axs[i].tick_params(axis='both', which='major', labelsize=10)
        #else: axs[i].set_yticklabels([])

    bbox_to_anchor = (1,-.2) if is_binary else (1,-.2)
    plt.legend(fancybox=True, bbox_to_anchor=bbox_to_anchor, ncol=4)

    type = 'binary' if is_binary else 'regression'
    filename = folder + '/calibration_' + type + '.pdf'
    plt.savefig(filename, dpi=1000, bbox_inches="tight")
    #plt.show()
    plt.clf()

# # # CONSISTENCY # # #

def plot_consistency(results, is_binary, algorithms, datasets, folder='figures', idx=42):
    fig, axs = plt.subplots(1, 5, figsize=(20, 3))    

    for i, dataset in enumerate(datasets):
        max_depths = sorted(results[dataset][list(algorithms.keys())[0]].keys())
        for j, name in enumerate(algorithms.keys()):
            std= np.array([results[dataset][name][max_depth][0][idx] for max_depth in max_depths])
            axs[i].plot(max_depths, std, label=name, linestyle=linestyles[j], linewidth=5)
        if i == 0: axs[i].set_ylabel('Uncertainty')

        axs[i].tick_params(axis='both', which='major', labelsize=10)
        axs[i].set_xlabel('Depth')
        axs[i].set_title(dataset)

    plt.legend(fancybox=True, bbox_to_anchor=(1,-.2), ncol=4)
    type = 'binary' if is_binary else 'regression'
    filename = folder + '/consistency_' + type + '.pdf'
    plt.savefig(filename, dpi=1000, bbox_inches="tight")
    #plt.show()
    plt.clf()

# # # ABSTENTION # # #

def plot_abstention(results, algorithms, datasets, percentiles, metrics, metric_name='Statistical Parity', folder='figure'):
    fig, axs = plt.subplots(1, len(datasets), figsize=(20, 3))

    for i, dataset in enumerate(datasets):
        instance = results[dataset]['instance']
        for j, algo_name in enumerate(algorithms):
            output = results[dataset][algo_name]
            metric_values = []
            for percentile in percentiles:
                std = output['std']
                include = std <= np.percentile(std, percentile)
                metric_value = metrics[metric_name](
                    output['pred'][include], instance['y_test'][include], instance['group_test'][include] == 1, run_checks=False
                )
                metric_values.append(metric_value)
            # increase line thickness
            axs[i].plot(1-percentiles/100, metric_values, label=algo_name, linestyle=linestyles[j], linewidth=3)
        if i == 0: axs[i].set_ylabel(metric_name)
        axs[i].tick_params(axis='both', which='major', labelsize=10)
        axs[i].set_title(dataset)
        axs[i].set_xlabel('Abstention Rate')

    plt.legend(fancybox=True, bbox_to_anchor=(1,-.2), ncol=4)
    filename = f'{folder}/abstention_{metric_name}.pdf'
    plt.savefig(filename, dpi=1000, bbox_inches="tight")
    #plt.show()
    plt.clf()

# # # REGRESSION FAIRNESS # # #

def plot_regression_fairness(results, datasets, algorithms, folder='figures'):
    fig, axs = plt.subplots(1, len(datasets), figsize=(20, 3))

    for dataset_num, dataset in enumerate(datasets):
        instance = results[dataset]['instance']
        min_z, max_z = min(instance['y_test']), max(instance['y_test'])
        z = np.linspace(min_z, max_z, 1000)

        colors = ['blue', 'red', 'green', 'orange', 'purple', 'brown', 'pink', 'gray', 'olive', 'cyan']
        for i, algorithm in enumerate(algorithms):
            output = results[dataset][algorithm]
            cdfs = []
            for j, group in enumerate(np.unique(instance['group_test'])):
                cdf = compute_cdf(output, z, instance['group_test'] == group)
                label = algorithm if j == 0 else None
                axs[dataset_num].plot(z, cdf, label=label, color=colors[i], linestyle=linestyles[j])
                cdfs += [cdf]
        axs[dataset_num].tick_params(axis='both', which='major', labelsize=10)
        axs[dataset_num].set_title(dataset)
        axs[dataset_num].set_xlabel(r'$y$')
        if dataset_num == 0:
            axs[dataset_num].set_ylabel('CDF')

    plt.legend(fancybox=True, bbox_to_anchor=(1,-.2), ncol=len(algorithms))
    plt.savefig(f'{folder}/regression_fairness.pdf', dpi=1000, bbox_inches='tight')
    #plt.show()
    plt.clf()


import scipy.stats
import torch
import numpy as np
import scipy
import xgboost as xgb
from torch.utils.data import DataLoader, Dataset

seed = 0
rng = np.random.default_rng(seed=seed)

def normal_nll(pred, dtrain):
    y = dtrain.get_label().reshape(-1, 2)[:,0]
    mu, sigma= pred[:,0], pred[:,1]
    # L = 2*np.log(sigma) + 0.5 * ((mu - y) / sigma) ** 2

    dLdmu = (mu- y) / sigma**2
    dLdsigma = -(mu- y)**2 / sigma**3 + 2 * sigma
    dLdmu2 = 1 / sigma**2
    dLdsigma2 = 3 * (mu - y)**2 / sigma**4 + 2

    grad = np.column_stack([dLdmu, dLdsigma]).reshape(-1, 1)

    hess = np.column_stack([dLdmu2, dLdsigma2]).reshape(-1, 1)

    return grad, hess

def beta_nll(pred, dtrain):
    y = dtrain.get_label().reshape(-1, 2)[:,0]
    mu, sigma= pred[:,0], pred[:,1]
    beta = .5
    weighting = sigma**(2*beta)
    # L = (2*np.log(sigma) + 0.5 * ((mu - y) / sigma) ** 2) * weighting

    dLdmu = (mu- y) / sigma**2 * weighting
    dLdsigma = (-(mu- y)**2 / sigma**3 + 2 * sigma) * weighting
    dLdmu2 = 1 / sigma**2 * weighting
    dLdsigma2 = (3 * (mu - y)**2 / sigma**4 + 2) * weighting

    grad = np.column_stack([dLdmu, dLdsigma]).reshape(-1, 1)

    hess = np.column_stack([dLdmu2, dLdsigma2]).reshape(-1, 1)

    return grad, hess
    
def faithful_nll(pred, dtrain):
    y = dtrain.get_label().reshape(-1, 2)[:,0]
    mu, sigma= pred[:,0], pred[:,1]
    # L = 2*np.log(sigma) + 0.5 * ((mu - y) / sigma) ** 2

    dLdmu = (mu- y)
    dLdsigma = -(mu- y)**2 / sigma**3 + 2 * sigma
    dLdmu2 = np.ones_like(dLdmu)
    dLdsigma2 = 3 * (mu - y)**2 / sigma**4 + 2

    grad = np.column_stack([dLdmu, dLdsigma]).reshape(-1, 1)

    hess = np.column_stack([dLdmu2, dLdsigma2]).reshape(-1, 1)

    return grad, hess


xgb_multi_losses = {
    'Normal NLL' : normal_nll,
    r'$\beta$-NLL' : beta_nll,
    'Faithful NLL' : faithful_nll,
}

def xgb_train(loss_name, xgb_instance):
    y = np.column_stack([xgb_instance['y_train'], np.zeros_like(xgb_instance['y_train'])])
    dtrain = xgb.DMatrix(xgb_instance['X_train'], label=y)
    bst = xgb.train(
        {
            'tree_method': 'hist',
            'max_depth' : xgb_instance['max_depth'],
            'disable_default_eval_metric' : True,
            'multi_strategy' : 'multi_output_tree',
        },
        dtrain = dtrain,
        num_boost_round = xgb_instance['num_boost_round'],
        obj = xgb_multi_losses[loss_name],
    )
    return lambda x: bst.predict(xgb.DMatrix(x)).squeeze()

class Model:
    def __init__(self, loss_name=None, instance=None):
        if instance is None:
            instance = {'max_depth' : 6, 'num_boost_round' : 10}
        self.max_depth = instance['max_depth']
        self.num_boost_round = instance['num_boost_round']
        self.loss_name = loss_name
        self.model = None
        if loss_name == 'classifier':
            self.model = xgb.XGBClassifier(max_depth=self.max_depth, n_estimators=self.num_boost_round)
        elif loss_name == 'regression':
            self.model = xgb.XGBRegressor(max_depth=self.max_depth, n_estimators=self.num_boost_round)
    
    def fit(self, X, y, sample_weight=None):
        if self.loss_name in ['classifier', 'regression']:
            self.model.fit(X, y, sample_weight=sample_weight)
        else:
            self.model = xgb_train(self.loss_name, {'X_train': X, 'y_train': y, 'max_depth': self.max_depth, 'num_boost_round': self.num_boost_round})
    
    def predict(self, X):
        if self.loss_name in ['classifier', 'regression']:
            return self.model.predict(X)
        return self.model(X)
    
    def predict_proba(self, X):
        return self.model.predict_proba(X)



import numpy as np
import scipy

rng = np.random.default_rng()

def self_inconsistency_ensemble(instance, Model, B=10):
    # https://arxiv.org/pdf/2301.11562
    preds = []
    n = len(instance['X_train'])
    for i in range(B):
        indices = rng.choice(n, n, replace=True)
        X = instance['X_train'][indices]
        y = instance['y_train'][indices]
        # HERE
        model = Model('classifier', instance)
        model.fit(X, y)
        preds.append(model.predict(instance['X_test']))
    preds = np.array(preds)
    B1 = np.sum(preds, axis=0)
    B0 = B - B1
    sc = 2 * B0 * B1 / (B * (B - 1))
    mid1, mid2 = (B+1) // 2, B // 2
    max_val = 2 * mid1 * mid2 / (B * (B - 1))
    assert np.all(sc <= max_val), 'Self inconsistency score is greater than theoretical maximum.'
    mode = scipy.stats.mode(preds, axis=0)[0].squeeze()
    return {'pred': mode, 'std': sc}


from fairlearn.postprocessing import ThresholdOptimizer
import warnings
import xgboost as xgb

def threshold_algorithm(instance, constraints):
    assert constraints in ['demographic_parity', 'equalized_odds'], 'Invalid constraints'
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        pred = ThresholdOptimizer(
            estimator=xgb.XGBClassifier(), constraints=constraints, predict_method='predict', prefit=False
        ).fit(
            instance['X_train'], instance['y_train'], sensitive_features=instance['group_train']
        ).predict(
            instance['X_test'], sensitive_features=instance['group_test']
        )
    return {'pred' : pred}

def threshold_sp(instance):
    return threshold_algorithm(instance, 'demographic_parity')

def threshold_eo(instance):
    return threshold_algorithm(instance, 'equalized_odds')


import numpy as np

def baseline(instance, Model):
    is_binary = len(np.unique(instance['y_train'])) == 2
    # HERE
    model = Model('classifier' if is_binary else 'regression', instance)
    model.fit(instance['X_train'], instance['y_train'])
    pred = model.predict(instance['X_test'])
    return {'pred': pred}

def true(instance):
    return {'pred' : instance['y_test']}

def random(instance, Model):
    output = baseline(instance, Model)
    output['std'] = np.random.rand(len(instance['y_test']))/2 # Random values between 0 and 0.5
    return output


import numpy as np
import scipy

rng = np.random.default_rng()
def compute_p_value(binary_preds):
    nA = np.sum(binary_preds==0, axis=0)
    nB = np.sum(binary_preds==1, axis=0)
    p_values = np.array([scipy.stats.binomtest(nA[i], nA[i]+nB[i], p=0.5).pvalue for i in range(binary_preds.shape[1])])
    return p_values

def selective_ensemble(instance, Model, num_ensemble=10):
    preds = []
    n = len(instance['X_train'])
    for i in range(num_ensemble):
        indices = rng.choice(n, n, replace=True)
        X = instance['X_train'][indices]
        y = instance['y_train'][indices]
        # HERE
        model = Model('classifier', instance)
        model.fit(X, y)
        preds.append(model.predict(instance['X_test']))
    preds = np.array(preds)
    p_values = compute_p_value(preds)
    mode = scipy.stats.mode(preds, axis=0)[0].squeeze()
    return {'pred': mode, 'std': p_values}


import numpy as np
from .model_xgb import Model

from .baseline import baseline, true, random
from .ensemble import ensemble
from .selective_ensemble import selective_ensemble
from .self_inconsistency import self_inconsistency_ensemble
from .binomial_nll import binomial_nll
from .regression_nll import normal_nll, beta_nll, faithful_nll
from .threshold import threshold_sp, threshold_eo
from .reductions import gridsearch, exponentiated_gradient, moments
from .ft_transformer import FTTransformerModel

# # # # BINARY AND REGRESSION # # # #

algorithms = {
    'True' : true,
    'Baseline' : lambda instance : baseline(instance, Model),
    'Random' : lambda instance : random(instance, Model),
    'Ensemble' : lambda instance : ensemble(instance, Model),
    'Selective Ensemble' : lambda instance : selective_ensemble(instance, Model),
    'Self-(in)consistency' : lambda instance : self_inconsistency_ensemble(instance, Model),
    'Binomial NLL' : lambda instance : binomial_nll(instance, Model),
    'Normal NLL' : lambda instance : normal_nll(instance, Model),
    r'$\beta$-NLL' : lambda instance : beta_nll(instance, Model),
    'Faithful NLL' : lambda instance : faithful_nll(instance, Model),
    'Threshold Optimizer SP' : threshold_sp,
    'Threshold Optimizer EO' : threshold_eo,
}

for moment in moments:
    algorithms[f'Exponentiated Gradient {moment}'] = lambda instance, moment=moment : exponentiated_gradient(instance, moment)

for moment in moments:
    algorithms[f'Grid Search {moment}'] = lambda instance, moment=moment : gridsearch(instance, moment)

binary_uncertainty = ['Ensemble', 'Selective Ensemble', 'Self-(in)consistency', 'Binomial NLL']

binary_fairness = ['Baseline']
for algorithm in algorithms:
    if 'SP' in algorithm or 'EO' in algorithm or 'ERP' in algorithm:
        binary_fairness.append(algorithm)

binary_fairness += ['Random', 'Ensemble', 'Selective Ensemble', 'Self-(in)consistency', 'Binomial NLL']

regression_uncertainty = ['Ensemble', 'Normal NLL', r'$\beta$-NLL', 'Faithful NLL']

regression_fairness = ['True', 'Baseline']

for algorithm in algorithms:
    if 'Square' in algorithm or 'Absolute' in algorithm:
        regression_fairness.append(algorithm)

regression_fairness += ['Ensemble', 'Normal NLL', r'$\beta$-NLL', 'Faithful NLL']



import numpy as np
from ucimlrepo import fetch_ucirepo 

def get_communities_and_crimes():
    instance = {}
    # fetch dataset 
    dataset = fetch_ucirepo(id=183) 

    dataset.data.features = dataset.data.features.replace('?', np.nan)
    # Drop columns with more than 10% missing values
    dataset.data.features.dropna(thresh=int(dataset.data.features.shape[0]), axis=1, inplace=True)
    dataset.data.features['sensitive'] = np.array((dataset.data.features['racepctblack'] > .2).astype(int))
    

    instance['data'] = dataset.data
    # Drop rows with missing 

    instance['y'] = np.array(dataset.data.targets)
    instance['group'] = dataset.data.features['sensitive'].values
    instance['n'] = len(dataset.data.features)

    cols_to_drop = ['state', 'communityname', 'fold']
    X = np.array(dataset.data.features.drop(cols_to_drop, axis=1).values)
    # Replace '?' with NaN
    X = np.where(X == '?', np.nan, X)
    # Convert to float
    X = X.astype(float)
    instance['X'] = X
    return instance


